from typing import Optional, Type, Union


def to_bool(value: Union[str, int, bool]) -> bool:
    if isinstance(value, str):
        return value.lower() in ('true', 't', '1')
    return bool(value)


def to_int(value: Union[str, int], default: int = None) -> Optional[int]:
    try:
        return int(value)
    except (TypeError, ValueError):
        return default


def to_list(value: str, cast: Type = None, separator: str = ',') -> list[str]:
    result = []
    if isinstance(value, str):
        result = [v for v in value.split(separator) if v]
    return [cast(v) for v in result] if callable(cast) else result



import functools
import os
from contextlib import ContextDecorator
from datetime import date, datetime, time, timedelta, timezone, tzinfo
from typing import Optional, Union

import pytz
from asgiref.local import Local

__all__ = [
    'utc', 'get_fixed_timezone', 'get_default_timezone', 'get_current_timezone', 'get_current_timezone_name',
    'get_default_timezone_name', 'activate', 'deactivate', 'override', 'localtime', 'localdate', 'now', 'is_aware', 'is_naive',
    'make_aware', 'make_naive', 'datetime', 'date', 'time', 'timezone', 'timedelta', 'tzinfo'
]

AnyDate = Union[datetime, date, time]

utc = pytz.utc


def get_fixed_timezone(offset: Union[timedelta, int]):
    if isinstance(offset, timedelta):
        offset = offset.total_seconds() // 60
    sign = '-' if offset < 0 else '+'
    hhmm = '%02d%02d' % divmod(abs(offset), 60)
    return timezone(timedelta(minutes=offset), sign + hhmm)


@functools.lru_cache()
def get_default_timezone():
    return pytz.timezone(os.environ.get('TIME_ZONE', 'Europe/Moscow'))


def get_default_timezone_name():
    return _get_timezone_name(get_default_timezone())


_active = Local()


def get_current_timezone():
    return getattr(_active, 'value', get_default_timezone())


def get_current_timezone_name():
    return _get_timezone_name(get_current_timezone())


def _get_timezone_name(timezone):
    return timezone.tzname(None) or str(timezone)


# Timezone selection functions.


def activate(timezone):
    if isinstance(timezone, tzinfo):
        _active.value = timezone
    elif isinstance(timezone, str):
        _active.value = pytz.timezone(timezone)
    else:
        raise ValueError(f'Invalid timezone: {timezone}')


def deactivate():
    if hasattr(_active, 'value'):
        del _active.value


class override(ContextDecorator):
    def __init__(self, timezone):
        self.timezone = timezone

    def __enter__(self):
        self.old_timezone = getattr(_active, 'value', None)
        if self.timezone is None:
            deactivate()
        else:
            activate(self.timezone)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.old_timezone is None:
            deactivate()
        else:
            _active.value = self.old_timezone


# Utilities

def localtime(value=None, timezone=None):
    if value is None:
        value = now()
    if timezone is None:
        timezone = get_current_timezone()
    if is_naive(value):
        raise ValueError('localtime() cannot be applied to a naive datetime')
    return value.astimezone(timezone)


def localdate(value=None, timezone=None):
    return localtime(value, timezone).date()


def now():
    return datetime.utcnow().replace(tzinfo=utc)


def is_aware(value):
    return value.utcoffset() is not None


def is_naive(value):
    return value.utcoffset() is None


def make_aware(value, tz=None, is_dst=None):
    if tz is None:
        tz = get_current_timezone()
    if _is_pytz_zone(tz):
        return tz.localize(value, is_dst=is_dst)
    else:
        if is_aware(value):
            raise ValueError(f'make_aware expects a naive datetime, got {value}')
        return value.replace(tzinfo=timezone)


def make_naive(value, timezone=None):
    if timezone is None:
        timezone = get_current_timezone()
    if is_naive(value):
        raise ValueError('make_naive() cannot be applied to a naive datetime')
    return value.astimezone(timezone).replace(tzinfo=None)


def parse(value: str, format: str, timezone: tzinfo = None) -> Optional[datetime]:
    return make_aware(datetime.strptime(value, format), timezone)


def has_overlap(start1: AnyDate, end1: AnyDate, start2: AnyDate, end2: AnyDate) -> bool:
    return max(start1, start2) <= min(end1, end2)


def _is_pytz_zone(tz):
    return isinstance(tz, (pytz.tzinfo.BaseTzInfo, pytz._FixedOffset))


def _datetime_ambiguous_or_imaginary(dt, tz):
    if _is_pytz_zone(tz):
        try:
            tz.utcoffset(dt)
        except (pytz.AmbiguousTimeError, pytz.NonExistentTimeError):
            return True
        else:
            return False

    return tz.utcoffset(dt.replace(fold=not dt.fold)) != tz.utcoffset(dt)



from typing import Any, Callable, Optional, Type, Union


class ClassPropertyDescriptor:
    def __init__(self, fget: Union[classmethod, staticmethod], fset: Optional[Union[classmethod, staticmethod]] = None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj: Any, cls: Type = None):
        if cls is None:
            cls = type(obj)
        return self.fget.__get__(obj, cls)()

    def __set__(self, obj: Any, value: Any):
        if not self.fset:
            raise AttributeError('can\'t set attribute')
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func: Union[classmethod, staticmethod, Callable]):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


def classproperty(func: Union[classmethod, staticmethod, Callable]):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)
    return ClassPropertyDescriptor(func)



import time
from typing import Callable


def timeit(method: Callable):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        print(f'{method.__name__}  {(te - ts) * 1000} ms')
        return result

    return timed



import datetime
import decimal
import json
import uuid
from json import JSONDecodeError, JSONDecoder  # noqa
from typing import Any, IO

from dimtim.interfaces import Serializable
from dimtim.utils import duration, timezone

__all__ = [
    'dump', 'dumps', 'load', 'loads', 'JSONEncoder', 'JSONDecoder', 'JSONDecodeError'
]

compat_transforms = {}

try:
    from django.utils.functional import Promise

    compat_transforms[Promise] = lambda o: str(o)
except ImportError:
    pass

try:
    from pydantic import BaseModel

    compat_transforms[BaseModel] = lambda o: o.dict()
except ImportError:
    pass


class JSONEncoder(json.JSONEncoder):
    def __init__(self, **kwargs):
        kwargs.setdefault('ensure_ascii', False)
        super().__init__(**kwargs)

    def default(self, o):
        if isinstance(o, datetime.datetime):
            r = o.isoformat()
            if o.microsecond:
                r = r[:23] + r[26:]
            if r.endswith('+00:00'):
                r = r[:-6] + 'Z'
            return r

        if isinstance(o, datetime.date):
            return o.isoformat()

        if isinstance(o, datetime.time):
            if timezone.is_aware(o):
                raise ValueError('JSON can\'t represent timezone-aware times.')
            r = o.isoformat()
            if o.microsecond:
                r = r[:12]
            return r

        if isinstance(o, datetime.timedelta):
            return duration.duration_iso_string(o)

        if isinstance(o, (decimal.Decimal, uuid.UUID)):
            return str(o)

        if isinstance(o, Serializable):
            return o.serialize()

        for t, fn in compat_transforms.items():
            if isinstance(o, t):
                return fn(o)

        return super().default(o)


def load(fp: IO, **kwargs) -> None:
    return json.load(fp, **kwargs)


def loads(data: str, **kwargs) -> Any:
    return json.loads(data, **kwargs)


def dump(obj: Any, fp: IO, **kwargs) -> None:
    kwargs.setdefault('cls', JSONEncoder)
    return json.dump(obj, fp, **kwargs)


def dumps(obj: Any, **kwargs) -> str:
    kwargs.setdefault('cls', JSONEncoder)
    return json.dumps(obj, **kwargs)



import re
import unicodedata
from typing import Optional


def slugify(value: str, allow_unicode: bool = False):
    if allow_unicode:
        value = unicodedata.normalize('NFKC', str(value))
    else:
        value = unicodedata.normalize('NFKD', str(value)).encode('ascii', 'ignore').decode('ascii')
    return re.sub(r'[-\s]+', '-', re.sub(r'[^\w\s-]', '', value.lower()).strip())


def strip_markdown(value: str):
    try:
        from bs4 import BeautifulSoup
        import markdown
    except ImportError:
        raise ModuleNotFoundError(f'Mthod "{strip_markdown.__name__}" requred "beautifulsoup4" and "markdown" libraries')

    return BeautifulSoup(markdown.markdown(value), 'lxml').text


def to_snake_case(text: str):
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', re.sub('(.)([A-Z][a-z]+)', r'\1_\2', text)).lower()


def to_camel_case(text: str):
    return ''.join(word.capitalize() for word in re.split(r'[ ._-]', text, re.MULTILINE))


KILOBYTE = 1024
MEGABYTE = 1048576
GIGABYTE = 1073741824
TERABYTE = 1099511627776
PETABYTE = 1125899906842624


def format_size(size: int):
    if size < KILOBYTE:
        return f'{size} b'
    if size < MEGABYTE:
        return f'{size / KILOBYTE:.2f} Kb'
    if size < GIGABYTE:
        return f'{size / MEGABYTE:.2f} Mb'
    if size < TERABYTE:
        return f'{size / GIGABYTE:.2f} Gb'
    if size < PETABYTE:
        return f'{size / TERABYTE:.2f} Tb'


def normalize_email(email: str) -> Optional[str]:
    try:
        email_name, domain_part = email.strip().rsplit('@', 1)
    except (ValueError, AttributeError):
        return None
    return '@'.join([email_name.lower(), domain_part.lower()])


def normalize_username(username: str) -> str:
    return unicodedata.normalize('NFKC', username) if username else username


def pluralize(value: int, one: str, two: str, five: str) -> str:
    ones = value % 10
    tens = value % 100
    return one if (ones == 1 and tens != 11) else two if (2 <= ones <= 4 and (tens < 10 or tens >= 20)) else five



import datetime


def _get_duration_components(duration: datetime.timedelta):
    minutes = duration.seconds // 60
    seconds = duration.seconds % 60

    hours = minutes // 60
    minutes = minutes % 60

    return duration.days, hours, minutes, seconds, duration.microseconds


def duration_string(duration: datetime.timedelta):
    days, hours, minutes, seconds, microseconds = _get_duration_components(duration)

    string = f'{hours:02d}:{minutes:02d}:{seconds:02d}'
    if days:
        string = f'{days} {string}'
    if microseconds:
        string += f'.{microseconds:06d}'

    return string


def duration_iso_string(duration: datetime.timedelta):
    if duration < datetime.timedelta(0):
        sign = '-'
        duration *= -1
    else:
        sign = ''

    days, hours, minutes, seconds, microseconds = _get_duration_components(duration)
    ms = f'.{microseconds:06d}' if microseconds else ""
    return f'{sign}P{days}DT{hours:02d}H{minutes:02d}M{seconds:02d}{ms}S'


def duration_microseconds(delta):
    return (24 * 60 * 60 * delta.days + delta.seconds) * 1000000 + delta.microseconds



import random
import secrets
import string
from typing import Sequence

from dimtim.utils.data.wordlists import ADJECTIVES, NOUNS

DEFAULT_CHARSET = string.ascii_letters + string.digits
PASSWORD_CHARSET = DEFAULT_CHARSET + string.punctuation
UPPER_CHARSET = string.ascii_uppercase + string.digits
HEX_CHARSET = string.hexdigits


def get_random_string(length: int, charset: str = DEFAULT_CHARSET):
    return ''.join(secrets.choice(charset) for _ in range(length))


def string4():
    return get_random_string(4, DEFAULT_CHARSET)


def string8():
    return get_random_string(8, DEFAULT_CHARSET)


def string16():
    return get_random_string(16, DEFAULT_CHARSET)


def string32():
    return get_random_string(32, DEFAULT_CHARSET)


def string64():
    return get_random_string(64, DEFAULT_CHARSET)


def hex4():
    return get_random_string(4, HEX_CHARSET)


def hex8():
    return get_random_string(8, HEX_CHARSET)


def hex16():
    return get_random_string(16, HEX_CHARSET)


def hex32():
    return get_random_string(32, HEX_CHARSET)


def hex64():
    return get_random_string(64, HEX_CHARSET)


def string8upper():
    return get_random_string(8, UPPER_CHARSET)


def username(underscores: bool = True):
    if underscores:
        return f'{random.choice(ADJECTIVES)}_{random.choice(ADJECTIVES)}_{random.choice(NOUNS)}'
    return f'{random.choice(ADJECTIVES).capitalize()}{random.choice(ADJECTIVES).capitalize()}{random.choice(NOUNS).capitalize()}'


def choice(variants: Sequence):
    return random.choice(variants)



import sys
from importlib import import_module


def safe_import(module: str, silence: bool = False, reraise: bool = False):
    try:
        return import_module(module)
    except ImportError as e:
        if not silence:
            sys.stdout.write(f'Module {module} importing error: {e}')
        if reraise:
            raise e



ADJECTIVES = [
    'different', 'important', 'every', 'large', 'available', 'popular', 'able', 'basic', 'known', 'various', 'difficult',
    'several', 'united', 'historical', 'hot', 'useful', 'mental', 'scared', 'additional', 'emotional', 'old', 'political',
    'similar', 'healthy', 'financial', 'medical', 'traditional', 'federal', 'entire', 'strong', 'actual', 'significant',
    'successful', 'electrical', 'expensive', 'pregnant', 'intelligent', 'interesting', 'poor', 'happy', 'responsible',
    'cute', 'helpful', 'recent', 'willing', 'nice', 'wonderful', 'impossible', 'serious', 'huge', 'rare', 'technical',
    'typical', 'competitive', 'critical', 'electronic', 'immediate', 'aware', 'educational', 'environmental', 'global',
    'legal', 'relevant', 'accurate', 'capable', 'dangerous', 'dramatic', 'efficient', 'powerful', 'foreign', 'hungry',
    'practical', 'psychological', 'severe', 'suitable', 'numerous', 'sufficient', 'unusual', 'consistent', 'cultural',
    'existing', 'famous', 'pure', 'afraid', 'obvious', 'careful', 'latter', 'unhappy', 'acceptable', 'aggressive', 'boring',
    'distinct', 'eastern', 'logical', 'reasonable', 'strict', 'administrative', 'automatic', 'civil', 'former', 'massive',
    'southern', 'unfair', 'visible', 'alive', 'angry', 'desperate', 'exciting', 'friendly', 'lucky', 'realistic', 'sorry',
    'ugly', 'unlikely', 'anxious', 'comprehensive', 'curious', 'impressive', 'informal', 'inner', 'pleasant', 'sexual',
    'sudden', 'terrible', 'unable', 'weak', 'wooden', 'asleep', 'confident', 'conscious', 'decent', 'embarrassed', 'guilty',
    'lonely', 'mad', 'nervous', 'odd', 'remarkable', 'substantial', 'suspicious', 'tall', 'tiny'
]

NOUNS = [
    'people', 'history', 'way', 'art', 'world', 'information', 'map', 'two', 'family', 'government', 'health', 'system',
    'computer', 'meat', 'year', 'thanks', 'music', 'person', 'reading', 'method', 'data', 'food', 'understanding', 'theory',
    'law', 'bird', 'literature', 'problem', 'software', 'control', 'knowledge', 'power', 'ability', 'economics', 'love',
    'internet', 'television', 'science', 'library', 'nature', 'fact', 'product', 'idea', 'temperature', 'investment', 'area',
    'society', 'activity', 'story', 'industry', 'media', 'thing', 'oven', 'community', 'definition', 'safety', 'quality',
    'development', 'language', 'management', 'player', 'variety', 'video', 'week', 'security', 'country', 'exam', 'movie',
    'organization', 'equipment', 'physics', 'analysis', 'policy', 'series', 'thought', 'basis', 'boyfriend', 'direction',
    'strategy', 'technology', 'army', 'camera', 'freedom', 'paper', 'environment', 'child', 'instance', 'month', 'truth',
    'marketing', 'university', 'writing', 'article', 'department', 'difference', 'goal', 'news', 'audience', 'fishing',
    'growth', 'income', 'marriage', 'user', 'combination', 'failure', 'meaning', 'medicine', 'philosophy', 'teacher',
    'communication', 'night', 'chemistry', 'disease', 'disk', 'energy', 'nation', 'road', 'role', 'soup', 'advertising',
    'location', 'success', 'addition', 'apartment', 'education', 'math', 'moment', 'painting', 'politics', 'attention',
    'decision', 'event', 'property', 'shopping', 'student', 'wood', 'competition', 'distribution', 'entertainment', 'office',
    'population', 'president', 'unit', 'category', 'cigarette', 'context', 'introduction', 'opportunity', 'performance',
    'driver', 'flight', 'length', 'magazine', 'newspaper', 'relationship', 'teaching', 'cell', 'dealer', 'finding', 'lake',
    'member', 'message', 'phone', 'scene', 'appearance', 'association', 'concept', 'customer', 'death', 'discussion',
    'housing', 'inflation', 'insurance', 'mood', 'woman', 'advice', 'blood', 'effort', 'expression', 'importance', 'opinion',
    'payment', 'reality', 'responsibility', 'situation', 'skill', 'statement', 'wealth', 'application', 'city', 'county',
    'depth', 'estate', 'foundation', 'grandmother', 'heart', 'perspective', 'photo', 'recipe', 'studio', 'topic',
    'collection', 'depression', 'imagination', 'passion', 'percentage', 'resource', 'setting', 'ad', 'agency', 'college',
    'connection', 'criticism', 'debt', 'description', 'memory', 'patience', 'secretary', 'solution', 'administration',
    'aspect', 'attitude', 'director', 'personality', 'psychology', 'recommendation', 'response', 'selection', 'storage',
    'version', 'alcohol', 'argument', 'complaint', 'contract', 'emphasis', 'highway', 'loss', 'membership', 'possession',
    'preparation', 'steak', 'union', 'agreement', 'cancer', 'currency', 'employment', 'engineering', 'entry', 'interaction',
    'mixture', 'preference', 'region', 'republic', 'tradition', 'virus', 'actor', 'classroom', 'delivery', 'device',
    'difficulty', 'drama', 'election', 'engine', 'football', 'guidance', 'hotel', 'owner', 'priority', 'protection',
    'suggestion', 'tension', 'variation', 'anxiety', 'atmosphere', 'awareness', 'bath', 'bread', 'candidate', 'climate',
    'comparison', 'confusion', 'construction', 'elevator', 'emotion', 'employee', 'employer', 'guest', 'height',
    'leadership', 'mall', 'manager', 'operation', 'recording', 'sample', 'transportation', 'charity', 'cousin', 'disaster',
    'editor', 'efficiency', 'excitement', 'extent', 'feedback', 'guitar', 'homework', 'leader', 'mom', 'outcome',
    'permission', 'presentation', 'promotion', 'reflection', 'refrigerator', 'resolution', 'revenue', 'session', 'singer',
    'tennis', 'basket', 'bonus', 'cabinet', 'childhood', 'church', 'clothes', 'coffee', 'dinner', 'drawing', 'hair',
    'hearing', 'initiative', 'judgment', 'lab', 'measurement', 'mode', 'mud', 'orange', 'poetry', 'police', 'possibility',
    'procedure', 'queen', 'ratio', 'relation', 'restaurant', 'satisfaction', 'sector', 'signature', 'significance', 'song',
    'tooth', 'town', 'vehicle', 'volume', 'wife', 'accident', 'airport', 'appointment', 'arrival', 'assumption', 'baseball',
    'chapter', 'committee', 'conversation', 'database', 'enthusiasm', 'error', 'explanation', 'farmer', 'gate', 'girl',
    'hall', 'historian', 'hospital', 'injury', 'instruction', 'maintenance', 'manufacturer', 'meal', 'perception', 'pie',
    'poem', 'presence', 'proposal', 'reception', 'replacement', 'revolution', 'river', 'son', 'speech', 'tea', 'village',
    'warning', 'winner', 'worker', 'writer', 'assistance', 'breath', 'buyer', 'chest', 'chocolate', 'conclusion',
    'contribution', 'cookie', 'courage', 'dad', 'desk', 'drawer', 'establishment', 'examination', 'garbage', 'grocery',
    'honey', 'impression', 'improvement', 'independence', 'insect', 'inspection', 'inspector', 'king', 'ladder', 'menu',
    'penalty', 'piano', 'potato', 'profession', 'professor', 'quantity', 'reaction', 'requirement', 'salad', 'sister',
    'supermarket', 'tongue', 'weakness', 'wedding', 'affair', 'ambition', 'analyst', 'apple', 'assignment', 'assistant',
    'bathroom', 'bedroom', 'beer', 'birthday', 'celebration', 'championship', 'cheek', 'client', 'consequence', 'departure',
    'diamond', 'dirt', 'ear', 'fortune', 'friendship', 'funeral', 'gene', 'girlfriend', 'hat', 'indication', 'intention',
    'lady', 'midnight', 'negotiation', 'obligation', 'passenger', 'pizza', 'platform', 'poet', 'pollution', 'recognition',
    'reputation', 'shirt', 'sir', 'speaker', 'stranger', 'surgery', 'sympathy', 'tale', 'throat', 'trainer', 'uncle', 'youth'
]



import hashlib
import hmac
from typing import Union


def md5(val: Union[str, bytes]):
    return hashlib.md5(val if isinstance(val, bytes) else str(val).encode()).hexdigest()


def sha1(val: Union[str, bytes]):
    return hashlib.sha1(val if isinstance(val, bytes) else str(val).encode()).hexdigest()


def secret_hmac(key: str, message: str):
    return hmac.digest(key.encode('utf8'), str(message).encode('utf8'), 'sha1').hex()



import os
from io import BytesIO
from typing import Union


def noop(*args, **kwargs):  # noqa
    pass


def fileext(name: str):
    return os.path.splitext(name)[-1]


def named_bytesio(name: str, content: Union[str, bytes] = None):
    bio = BytesIO(content)
    bio.name = name
    return bio
